/*********************************************************************
THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND, AND
ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH (JAIR)
AND JAIR'S PUBLISHERS AND DISTRIBUTERS, DISCLAIM ANY AND ALL WARRANTIES,
INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY WARRANTIES OR NON
INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND RESPONSIBILITY FOR
USE OF THIS SOURCE CODE AND NEITHER THE AUTHOR NOR JAIR, NOR JAIR'S
PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR DAMAMAGES OF ANY KIND
RESULTING FROM ITS USE.  Without limiting the generality of the
foregoing, neither the author, nor JAIR, nor JAIR's publishers and
distributors, warrant that the source code will be error-free, will
operate without interruption, or will meet the needs of the user.
**********************************************************************/

/* Copyright Geoffrey I. Webb 1995. */

#include "defns.i"
#include "types.i"
#include "extern.i"

#include <values.h>
#include <malloc.h>

static ItemCount *NullClassDist = NULL;

extern ItemCount *Weight;	/* Weight[i]  = current fraction of item i */

typedef struct CutRecS {
	Attribute	att;
	float		cut;
	Boolean		lte_branch;
	struct CutRecS  *next;
} CutRec;

CutRec *cuts = NULL;


/*************************************************************************/
/*								 	 */
/*		Add a cut to the stack of active cuts                    */
/*								 	 */
/*************************************************************************/

static PushCut(att, cut, lte_branch)
Attribute att;
float cut;
{   CutRec *rec;

    rec = (CutRec*)malloc(sizeof(CutRec));

    rec->att = att;
    rec->cut = cut;
    rec->lte_branch = lte_branch;
    rec->next = cuts;

    cuts = rec;
}


/*************************************************************************/
/*								 	 */
/*		pop a cut from the stack of active cuts                  */
/*								 	 */
/*************************************************************************/

static PopCut()
{   CutRec *old = cuts;

    cuts = cuts->next;

    free(old);
}



/****************************************************************************/
/* 									    */
/* Return the maximum value possible for this attribute at the current leaf */
/* 									    */
/****************************************************************************/

static float PartitionMax(att)
Attribute att;
{   CutRec *rec = cuts;
    float MaxThresh = MAXFLOAT;

    while (rec != NULL) {
	if (rec->att == att) {
	    if (rec->lte_branch && rec->cut < MaxThresh) {
		MaxThresh = rec->cut;
	    }
	}

	rec = rec->next;
    }

    return MaxThresh;
}



/****************************************************************************/
/* 									    */
/* Return the minimum value possible for this attribute at the current leaf */
/* 									    */
/****************************************************************************/

static float PartitionMin(att)
Attribute att;
{   CutRec *rec = cuts;
    float MinThresh = -MAXFLOAT;

    while (rec != NULL) {
	if (rec->att == att) {
	    if ((!rec->lte_branch) && rec->cut > MinThresh) {
		MinThresh = rec->cut;
	    }
	}

	rec = rec->next;
    }

    return MinThresh;
}



typedef struct RangeRecS {
	struct RangeRecS  *next;
	ItemNo Fp, Lp;
	ItemCount *weights;/* alloc dynamically */
	Description *items;
} RangeRec;

RangeRec *ranges = NULL;


/*************************************************************************/
/*								 	 */
/*		Add a range to the stack of active ranges                */
/*								 	 */
/*************************************************************************/

static PushRange(Fp, Lp)
ItemNo Fp, Lp;
{   RangeRec *rec;

    rec = (RangeRec*)malloc(sizeof(RangeRec));
    rec->weights = (ItemCount*)malloc((Lp+1)*sizeof(ItemCount));
    rec->items = (Description*)malloc((Lp+1)*sizeof(Description));

    rec->Fp = Fp;
    rec->Lp = Lp;
    memcpy(rec->weights, Weight, (Lp+1)*sizeof(ItemCount));
    memcpy(rec->items, Item, (Lp+1)*sizeof(Description));
    rec->next = ranges;

    ranges = rec;
}


/*************************************************************************/
/*								 	 */
/*		pop a range from the stack of active ranges              */
/*								 	 */
/*************************************************************************/

static PopRange()
{   RangeRec *old = ranges;

    ranges = ranges->next;

    free(old->items);
    free(old->weights);
    free(old);
}



static ItemCount *ItemCounts = NULL;



/*************************************************************************/
/*									 */
/*  Return true if this cut has a lower error estimate for another class */
/*  Set newclass to that class						 */
/*									 */
/*************************************************************************/
int AltCut(Att, MinThresh, MaxThresh, MinBranch, AccEst, newclass, NewThresh)
Attribute Att;
float MinThresh;
float MaxThresh;
Boolean MinBranch;
float *AccEst;
ClassNo *newclass;
float *NewThresh;
{   RangeRec *range = ranges;
    int found = false;
    int depth = 0;
    ItemNo Case;

/*printf("AltCut(%d, %3.1f, %3.1f, %d, %4.2f)\n", Att, MinThresh, MaxThresh, MinBranch, *AccEst);*/
    /* exit quickly if no alternatives are possible */
    if (MinThresh >= MaxThresh) return false;

    if (ItemCounts == NULL) {
	ItemCounts = (ItemCount*) calloc(MaxClass+1, sizeof(ItemCount));
    }

    while (range != NULL) {
	float CurMinThresh = MinThresh;
	float CurMaxThresh = MaxThresh;

	depth++;

	ForEach (Case, range->Fp, range->Lp) {
	    float ThisThresh = CVal(range->items[Case], Att);

	    if ( ThisThresh != Unknown &&
		 ThisThresh >= CurMinThresh &&
		 ThisThresh < CurMaxThresh
	       ) {
		ItemNo i;
		ItemCount total = 0.0;
		float ThisMin, ThisMax;

		if ( MinBranch ) {
		    ThisMin = MinThresh;
		    ThisMax = ThisThresh;
		}
		else {
		    ThisMin = ThisThresh;
		    ThisMax = MaxThresh;
		}

		ForEach (i, 0, MaxClass) {
		    ItemCounts[i] = 0.0;
		}

		ForEach (i, range->Fp, range->Lp) {
		    if ( CVal(range->items[i], Att) != Unknown &&
			 CVal(range->items[i], Att) > ThisMin &&
			 CVal(range->items[i], Att) <= ThisMax ) {
			ItemCounts[ Class(range->items[i]) ] += range->weights[i];
			total += range->weights[i];
		    }
		}

		if ( (total + 1.0)/(total + 2.0) <= *AccEst) {
		    /* too few cases to obtain a positive result             */
		    /* prevent more extreme thresholds from being considered */

		    if ( MinBranch ) {
			CurMinThresh = ThisThresh + Epsilon;
		    }
		    else {
			CurMaxThresh = ThisThresh;
		    }
		}
		else
		{

		    ForEach (i, 0, MaxClass) {
			float Est;

			Est = (ItemCounts[i] + 1.0) / (total + 2.0);

			if ( Est > *AccEst ) {
			    *AccEst = Est;
			    *newclass = i;
			    *NewThresh = ThisThresh;
			    found = true;
			}
		    }
		}
	    }
	}

	range = range->next;
    }

    return found;
}


#define MIN 1
#define MAX 2

GenLeaf(node, Fp, Lp)
Tree node;
ItemNo Fp, Lp;
{   int att;
    ClassNo oldclass;
    Tree Leaf();
    float AccEst;	/* estimate of error for the node */
    float Best;
    Boolean Add = 0;
    int AddAtt;
    float AddCut;
    ClassNo AddClass;

/*printf("SpecialiseTo(Fp=%d,Lp=%d)\n",Fp,Lp);*/
    oldclass = node->Leaf;

    /* calculate the error estimate */
    {   float count = 0.0;
	int c;

	ForEach (c, 0, MaxClass) {
	    count += node->ClassDist[c];
	}

	AccEst = (node->ClassDist[oldclass] + 1.0) / (count + 2.0);
    }

    /* allocate memory to NullClassDist if necessary */
    if (NullClassDist == NULL) {
	NullClassDist = (ItemCount *) calloc(MaxClass+1, sizeof(ItemCount));
    }

    Best = AccEst;

    /* find the best additional split to add */
    ForEach (att, 0, MaxAtt) {

	if ( SpecialStatus[att] == IGNORE ) continue;

	if ( MaxAttVal[att] ) {
	    /*  ignore discrete valued attribute  */
	}
	else {
	    /*  continuous attribute  */
	    float min, max;	/* minimum and maximum values for this att */
	    ItemNo i;

	    /* set min */
	    min = CVal(Item[Fp], att);

	    ForEach (i, Fp+1, Lp) {
		float v = CVal(Item[i], att);

		if ( v != Unknown &&
		     (min == Unknown || v < min) ) {
		    min = v;
		}
	    }

	    if (min != Unknown ) {
		ClassNo newclass;

		if (AltCut(att, PartitionMin(att), min-Epsilon, true, &Best, &newclass, &AddCut)) {
		    Add = MIN;
		    AddAtt = att;
		    AddClass = newclass;
		}
	    }

	    /* set max */
	    max = CVal(Item[Fp], att);

	    ForEach (i, Fp+1, Lp) {
		if (max == Unknown || CVal(Item[i], att) > max) max = CVal(Item[i], att);
	    }

	    if (max != Unknown) {
		ClassNo newclass;

		if (AltCut(att, max, PartitionMax(att), false, &Best, &newclass, &AddCut)) {
		    Add = MAX;
		    AddAtt = att;
		    AddClass = newclass;
		}
	    }
	}
    }
    
    if ( AddClass != oldclass ) {
	if (Add == MIN) {
	    ItemNo i;

	    /* initialise NullClassDist */
	    ForEach (i, 0, MaxClass) NullClassDist[i] = 0.0;

	    /* provide a minor weighting toward the new class */
	    NullClassDist[AddClass] = Epsilon;

	    Sprout(node, 2);

	    node->NodeType      = ThreshContin;
	    node->Tested        = AddAtt;
	    node->Cut           =
	    node->Lower         =
	    node->Upper         = AddCut;
	    node->Branch[1]	= Leaf(NullClassDist, AddClass, Epsilon, 0.0);

	    /* this is now the branch for the original class */
	    node = node->Branch[2] =  Leaf(node->ClassDist, oldclass, node->Items, node->Errors);
	}
	else if (Add == MAX) {
	    ItemNo i;

	    /* initialise NullClassDist */
	    ForEach (i, 0, MaxClass) NullClassDist[i] = 0.0;

	    /* provide a minor weighting toward the new class */
	    NullClassDist[AddClass] = Epsilon;

	    Sprout(node, 2);

	    node->NodeType   = ThreshContin;
	    node->Tested     = AddAtt;
	    node->Cut        =
	    node->Lower      =
	    node->Upper      = AddCut;
	    node->Branch[1]	=  Leaf(node->ClassDist, oldclass, node->Items, node->Errors);
	    node->Branch[2]	= Leaf(NullClassDist, AddClass, Epsilon, 0.0);

	    /* this is now the branch for the original class */
	    node = node->Branch[1];
	}
    }
}

GenAux(node, Fp, Lp)
Tree node;
ItemNo Fp, Lp;
{   ItemCount CountItems();
    ItemNo Group();

    if (node->NodeType == 0) {
	/* a leaf node */

	GenLeaf(node, Fp, Lp);
    }
    else {
	int i, v;
	ItemNo Kp, Ep;
	ItemCount Factor, KnownCases;
	Boolean CutPushed;

        PushRange(Fp, Lp);

	/* group the cases with unknown values for this attribute */
        Kp = Group(0, Fp, Lp, node) + 1;

	KnownCases = CountItems(Kp, Lp);

	ForEach(v, 1, node->Forks)
	{
	    if ( !MaxAttVal[node->Tested] ) {
		/* continuous attribute */
		PushCut(node->Tested, node->Cut, v == 1);
		CutPushed = true;
	    }
	    else CutPushed = false;

	    Ep = Group(v, Kp, Lp, node);

	    /* bug fix 10/4/96 - Kp changed to Fp to allow for inclusion of unknowns */
	    if ( KnownCases > 0.0 )
	    {
		Factor = CountItems(Kp, Ep) / KnownCases;
	    }
	    else
	    {
		Factor = 1E10;
	    }

	    ForEach(i, Fp, Kp-1)
	    {
		    Weight[i] *= Factor;
	    }

	    /*GenAux(node->Branch[i], Kp, Ep);*/
	    /* include unknowns */
	    GenAux(node->Branch[v], Fp, Ep);

	    /* restore weights for cases with unknown values for the attribute */
	    Group(0, Fp, Ep, node);
	    ForEach(i, Fp, Kp-1)
	    {
		    Weight[i] /= Factor;
	    }

	    if ( CutPushed ) {
		PopCut();
	    }
	}

        PopRange();
    }
}

Generalise(tree)
Tree tree;
{   
    InitialiseWeights();

    GenAux(tree, 0, MaxItem);
}
